"""Release [0.31.1].

Revision ID: 0.31.1
Revises: 4a3087070f4e
Create Date: 2023-01-13 16:29:44.076409

"""


# revision identifiers, used by Alembic.
revision = "0.31.1"
down_revision = "4a3087070f4e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
