"""Release [0.36.1].

Revision ID: 0.36.1
Revises: 728c6369cfaa
Create Date: 2023-03-24 09:55:43.989307

"""


# revision identifiers, used by Alembic.
revision = "0.36.1"
down_revision = "728c6369cfaa"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
