"""Release [0.21.0].

Revision ID: 0.21.0
Revises: d02b3d3464cf
Create Date: 2022-11-03 17:02:41.529907

"""


# revision identifiers, used by Alembic.
revision = "0.21.0"
down_revision = "d02b3d3464cf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
